#include<stdio.h>
#include<unistd.h>
#include<stdlib.h>
#include<string.h>
#include<sys/socket.h>
#include<netinet/in.h>
#include<arpa/inet.h>
#include"rr.h"
#include"rrparser.h"
#include"errors.h"
#include"zone.h"

/*
 * zone file format:
 * <#RR><RR0len><RR0><RR1len><RR1>...<RRnlen><RRn>
 * 
 * where #RR = n (=>counted w/o RR0) [uint_16]
 * 
 * where RR0 is special type T_ZONE with RDATA format:
 * <ORIGIN/ZONENAME(labelsstr)>
 * (ttl not meanable); RDATA len not checked to be
 * compatible, only first field read'd
 */

int
read_zone(int fd, zone_t *zone) {
  rr_t zonerr; int n, r;
  
  read(fd, &zone->rrs, 2);
  if ((r=read_raw_rr(fd, &zonerr))) return r;
  if (ntohs(zonerr.rrtype)!=T_ZONE) return XEINVAL;

  zone->origin=malloc(*zonerr.rdata+1);
  strncpy(zone->origin, zonerr.rdata, *zonerr.rdata+1);

  zone->soa=NULL;

  zone->rr=malloc(ntohs(zone->rrs)*sizeof(rr_t));
  for (n=0; n<ntohs(zone->rrs); n++) {
    if ((r=read_raw_rr(fd, &zone->rr[n]))) return r;
    if (ntohs(zone->rr[n].rrtype)==T_SOA) zone->soa=&zone->rr[n];
  }

  if(!zone->soa) return XENOENT;
  
  return XENONE;
}

int
dump_zone(int fd, zone_t *zone) {
  rr_t zonerr; int n, r;
  
  write(fd, &zone->rrs, 2);
  zonerr.name="";
  zonerr.rrtype=htons(T_ZONE);
  zonerr.rrclass=htons(1);
  zonerr.ttl=0;
  zonerr.rdlength=*zone->origin+1;
  zonerr.rdata=malloc(zonerr.rdlength);
  strncpy(zonerr.rdata, zone->origin, zonerr.rdlength);
  zonerr.rdlength=htons(zonerr.rdlength);
    
  if ((r=dump_raw_rr(fd, &zonerr))) return r;

  for (n=0; n<ntohs(zone->rrs); n++) {
    if ((r=dump_raw_rr(fd, &zone->rr[n]))) return r;
  }
  
  return XENONE;
}

int
dump_str_zone(char **l, zone_t *zone) {
  int m, n, ll; static char **x;
  x=malloc(sizeof(char *));
  
  labels2domain(zone->origin);
  *l=malloc(ll=(strlen(zone->origin)+2));
  sprintf(*l, "%s\n", zone->origin);
  
  for(n=0; n<ntohs(zone->rrs); n++) {
    if ((m=dump_str_rr(x, &zone->rr[n]))) return m;
    *l=realloc(*l, ll+=(strlen(*x)+1));
    sprintf(*l, "%s%s\n", *l, *x);
    free(*x);
  }

  return XENONE;
}

int
read_str_zone(FILE *f, zone_t *zone) {
  char x[256]; int n;
  zone->origin=malloc(257);
  
  fgets(zone->origin+1, 255, f);
  
  *zone->origin=strlen(zone->origin+2);
  domain2labels(zone->origin);
  
  zone->rr=NULL; zone->soa=NULL;
  for(n=0; fgets(x, 255, f); n++) {
    zone->rr=realloc(zone->rr, (n+1)*sizeof(rr_t));
    read_str_rr(x, &zone->rr[n]);
    if (ntohs(zone->rr[n].rrtype)==T_SOA) zone->soa=&zone->rr[n];
  }
  
  if (!zone->soa) return XENOENT;
  
  zone->rrs=htons(n);

  return XENONE;
}
